/**
 * @file 	HeuristicEvaluator.java
 * 
 * @date 	Aug 17, 2004
 * @author 	Alexandre Bouchard
 */

package ca.mcgill.rl.apps;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.PrintWriter;

import ca.mcgill.rl.ndp.Action;
import ca.mcgill.rl.ndp.EnvironmentSimulator;
import ca.mcgill.rl.ndp.HeuristicPolicy;
import ca.mcgill.rl.util.Configurable;
import ca.mcgill.rl.util.EuclideanLocation;
import ca.mcgill.rl.util.MultiInterval;
import ca.mcgill.rl.util.TypedProperties;

/**
 *
 */
public class HeuristicEvaluator implements Configurable
{
    /**
     * TODO:
     * Comment for <code>environment</code>
     */
    private EnvironmentSimulator environment; 
    
    /**
     * TODO:
     * Comment for <code>heuristicPolicy</code>
     */
    private HeuristicPolicy heuristicPolicy;
    
    /**
     * TODO:
     * Comment for <code>discountFactor</code>
     */
    private int discountFactor;
    
    /**
     * TODO:
     * Comment for <code>propertyPrefix</code>
     */
    private static String propertyPrefix = "HeuristicEvaluator.";   
    
    /**
     * TODO:
     * Comment for <code>experimentDirectory</code>
     */
    private File experimentDirectory;

    /**
     * TODO:
     * Comment for <code>numberOfGrids</code>
     */
    private int numberOfGrids;
    
    /**
     * TODO:
     * Comment for <code>trajectoryLength</code>
     */
    private int trajectoryLength;

    /**
     * TODO:
     * Comment for <code>trajectoryPerGridPoint</code>
     */
    private int trajectoryPerGridPoint;
    
    /**
     * TODO:
     * Comment for <code>stateSpace</code>
     */
    private MultiInterval stateSpace;
    
    /**
     * @param experimentDirectory
     */
    public HeuristicEvaluator(File experimentDirectory)
    {
        this.experimentDirectory = experimentDirectory;
    }    

    /**
     * 
     */
    public void startEvaluation()
    {
        try
        {
            PrintWriter out = new PrintWriter(new FileOutputStream(new File(experimentDirectory, "evaluations")));
            EuclideanLocation location = new EuclideanLocation(stateSpace.getDimensionality());
            for (int i = 0; i < stateSpace.getDimensionality(); i++)
            {
                location.setCoordinate(i, stateSpace.getLowerBound(i));
            }
            recursivePrintStatistics(out, location, stateSpace.getDimensionality() - 1);
            out.close();
        }
        catch (FileNotFoundException e)
        {
            // TODO Auto-generated catch block
            e.printStackTrace();
        }        
    }
    
    /**
     * @param out
     * @param i
     */
    private void recursivePrintStatistics(PrintWriter out, EuclideanLocation currentLocation, int currentDimension)
    {
        if (currentDimension == 0)
        {
            double increment = (stateSpace.getUpperBound(0) - stateSpace.getLowerBound(0)) / numberOfGrids;
            currentLocation.setCoordinate(0, stateSpace.getLowerBound(0));
            for (int i = 0; i <= numberOfGrids; i++)
            {
                double evaluation = evaluate(currentLocation);
                for (int j = 0; j < stateSpace.getDimensionality(); j++)
                {
                    out.print("" + currentLocation.getCoordinate(j) + " ");
                }
                out.print(evaluation);
                out.println();
                currentLocation.setCoordinate(0, currentLocation.getCoordinate(0) + increment);
            }
        }
        else
        {
            double increment = (stateSpace.getUpperBound(currentDimension) - stateSpace.getLowerBound(currentDimension)) / numberOfGrids;
            currentLocation.setCoordinate(currentDimension, stateSpace.getLowerBound(currentDimension));
            for (int i = 0; i <= numberOfGrids; i++)
            {
                recursivePrintStatistics(out, currentLocation, currentDimension - 1);
                currentLocation.setCoordinate(currentDimension, currentLocation.getCoordinate(currentDimension) + increment);
            }
        }
    }    
    
    /**
     * @param currentLocation
     * @return
     */
    private double evaluate(EuclideanLocation currentLocation)
    {
        double totalReward = 0;
        for (int k = 0; k < trajectoryPerGridPoint; k++)
        {
			environment.setState(currentLocation);
			double currentDiscount = 1;
			int j;
			for (j = 0; j < trajectoryLength && !environment.isAtTerminalState(); j++)
			{
				Action selectedAction = heuristicPolicy.selectAction(environment.getState(), environment.getPossibleActions());
				environment.generateTransition(selectedAction);
				totalReward = currentDiscount * environment.getReward() + totalReward;
				currentDiscount = currentDiscount * discountFactor;
			}
        }
        return (totalReward / (double) trajectoryPerGridPoint);
    }

    /**
     * @param properties
     */
    public void configure(TypedProperties properties)
    {
        trajectoryPerGridPoint = properties.getPropertyAsInt(propertyPrefix + "trajectoryPerGridPoint");
        numberOfGrids = properties.getPropertyAsInt(propertyPrefix + "numberOfGrids");     
        trajectoryLength = properties.getPropertyAsInt(propertyPrefix + "trajectoryLength");
        discountFactor = properties.getPropertyAsInt(propertyPrefix + "discountFactor");
        
        environment = (EnvironmentSimulator) properties.getPropertyAsObject(propertyPrefix + "environment");
        if (environment instanceof Configurable)
        {
            ((Configurable) environment).configure(properties);
        }
        heuristicPolicy = (HeuristicPolicy) properties.getPropertyAsObject(propertyPrefix + "heuristicPolicy");
        heuristicPolicy.setEnvironment(environment);
        if (heuristicPolicy instanceof Configurable)
        {
            ((Configurable) heuristicPolicy).configure(properties);
        }
        stateSpace = (MultiInterval) environment.getStateSpace();
    }
    
    /**
     * @param args
     */
    public static void main(String [] args)
    {
        if (args.length != 1)
        {
            throw new RuntimeException("You must specify the directory where the experiment parameters will be read and the results written.");
        }
        try
        {
            File experimentDirectory = new File(args[0]);
            HeuristicEvaluator evaluator = new HeuristicEvaluator(experimentDirectory);
            File propertyFile = new File(experimentDirectory, "evaluatorParameters.properties");
            TypedProperties properties = new TypedProperties();
            properties.load(new FileInputStream(propertyFile));
            evaluator.configure(properties);
            evaluator.startEvaluation();
        }
        catch (FileNotFoundException e)
        {
            // TODO Auto-generated catch block
            e.printStackTrace();
        }
        catch (IOException e)
        {
            // TODO Auto-generated catch block
            e.printStackTrace();
        }
    }
}
