/**
 * @file Approximator.java
 * 
 * @date Aug 7, 2004
 * @author Alexandre Bouchard
 */

package ca.mcgill.rl.ndp;

import ca.mcgill.rl.util.AffineVector;
import ca.mcgill.rl.util.RealValuedFunction;
import ca.mcgill.rl.util.State;

/**
 * An interface for parametrized, real-valued function approximators.
 * 
 * Formally, the subclasses of Approximator should represent maps z : R^n -> B(X),
 * where B(X) is the set of bounded functions on the state space X, evaluated at
 * some parameters vector. Hence they should also have some field p for keeping the 
 * current parameters of z (but an AffineVector is not necessarily the best type
 * to store it so this field is not included in this interface). We will call the
 * approximation function the bounded function z(p).
 */
public interface Approximator extends RealValuedFunction
{
	/**
	 * Increments the parameters by a given amount.
	 * 
	 * Should be equivalent to p := p + (p's type) parametersIncrement.
	 * @param	parametersIncrement The amount the parameters should be incremented
	 * 			in each dimension.
	 * @param multiplicativeConstant TODO
	 */
	public void updateParameters(AffineVector parametersIncrement);
	
	/**
	 * Evaluates the gradient of the approximation function at the given point in the
	 * state space.
	 * 
	 * Should return (grad z(p))(state).
	 * 
	 * @param	state The point in the state space where the gradient should be
	 * 			evaluated
	 */
	public AffineVector computeGradient(State state);
}
