/**
 * @file 	MountainCarEnvironment.java
 * 
 * @date 	Aug 8, 2004
 * @author 	Alexandre Bouchard
 */

package ca.mcgill.rl.ndp;

import java.util.ArrayList;
import java.util.List;
import java.util.Random;

import ca.mcgill.rl.util.DiscreteMeasure;
import ca.mcgill.rl.util.EuclideanLocation;
import ca.mcgill.rl.util.MultiInterval;
import ca.mcgill.rl.util.ProbabilityMeasure;
import ca.mcgill.rl.util.State;
import ca.mcgill.rl.util.StateSpace;

/**
 * The mountain car domain, as described in 
 * 
 * http://www-anw.cs.umass.edu/~rich/book/8/node9.html#figcarepisode
 */
public class MountainCarEnvironment implements ModelledEnvironmentSimulator
{
    /**
     * The current (position, velocity) of the car.
     */
    private EuclideanLocation currentState;
    
    /**
     * The current reward.
     */
    private double currentReward = 0;
    
    /**
     * The possible actions.
     */
    private List possibleActions = new ArrayList(2);
    
    /**
     * A time-seeded pseudo-random number generator.
     */
    private Random randomNumberGenerator = new Random(System.currentTimeMillis());
    
    /**
     * TODO:
     * Comment for <code>stateSpace</code>
     */
    private StateSpace stateSpace;
    
    /**
     * Create a new instance. Note that to complete initialization, the state should
     * be set (either with setState() or setStateRandomly()) before calling any
     * other method.
     */
    public MountainCarEnvironment()
    {
        // Two actions are possibles: move forward, move backward.
        possibleActions.add(new DiscreteAction(1));
        possibleActions.add(new DiscreteAction(-1));
        
        List lowerBounds = new ArrayList(2);
        lowerBounds.add(new Double(-1.2));
        lowerBounds.add(new Double(-0.07));
        List upperBounds = new ArrayList(2);
        upperBounds.add(new Double(0.5));
        upperBounds.add(new Double(0.07));
        stateSpace = new MultiInterval(lowerBounds, upperBounds);
    }

    /**
     * Returns the current (position, velocity) of the car.
     * 
     * Throws a RuntimeException if the state was not set.
     * 
     * @return The current (position, velocity) of the car.
     */
    public State getState()
    {
        if (currentState == null)
        {
            throw new RuntimeException("Uninitialized MountainCarEnvironment.");
        }
        return currentState;
    }

    /**
     * Returns the reward incurred by the last transition.
     * 
     * Throws a RuntimeException if the state was not set. Returns 0 if a transition
     * did not happened yet.
     * 
     * @return The reward incurred by the last transition.
     */
    public double getReward()
    {
        if (currentState == null)
        {
            throw new RuntimeException("Uninitialized MountainCarEnvironment.");
        }     
        return currentReward;
    }

    /**
     * Generate a transition given a selected action.
     * 
     * An action of 1 corresponds to "full throttle to the right" and an action of
     * -1 corresponds to "full throttle to the left". The new state is computed 
     * using newtownian movement equation, with a delta time of 1 and little g 
     * gravity constant 0.0025. The acceleration if +- 0.001, depending on the
     * direction it is applied. The car is on a hill of shape 3.Sin(3.x) and the
     * space is bounded in velocity in the interval [-0.07, 0.07] and in position
     * in the interval [-1.2, 0.5]. The goal state is to get at position 0.5. When 
     * the velocity boundary is reached, further acceleration in the direction of the
     * velocity will not increase the velocity. Reaching the space boundary -1.2 will
     * keep the position at -1.2 and set the velocity to 0 (the car "hits a wall").
     * 
     * Checks that the selected action is valid, else throws a RuntimeException.
     * 
     * @param selectedAction
     */
    public void generateTransition(Action selectedAction)
    {
        double actionValue = ((DiscreteAction) selectedAction).getValue();
        if (actionValue != 1 && actionValue != -1)
        {
            throw new RuntimeException("Illegal action selected in MountainCarEnvironment.");
        }
        double currentPosition = currentState.getCoordinate(0);
        double currentVelocity = currentState.getCoordinate(1);       
        
        // Compute the new velocity using:
        // delta velocity = (acceleration of the car - gravity accel.) * delta time
        double nextVelocity = currentVelocity + 0.001 * actionValue
			- 0.0025 * Math.cos((double) 3 * currentPosition);
        if (nextVelocity < -0.07) 
        {
            nextVelocity = -0.07;
        }
        if (nextVelocity > 0.07)
        {
            nextVelocity = 0.07;
        }
        
        currentReward = -1.0; //until goal is reached
        
        // Compute the new position using:
        // delta position = velocity * delta time
        double nextPosition = currentPosition + nextVelocity;
        if (nextPosition < -1.2) 
		{
		    nextPosition = -1.2; 
		    nextVelocity = 0.0;
		}
        if (nextPosition >= 0.5)
        {
            nextPosition = 0.5;
            currentReward = 0.0;
        }     
		
        currentState = new EuclideanLocation(2);
        currentState.setCoordinate(0, nextPosition);
        currentState.setCoordinate(1, nextVelocity);
    }

    /**
     * Returns the actions allowed from the current state, i.e. move forward, move
     * backward.
     * 
     * @return The actions allowed from the current state.
     */
    public List getPossibleActions()
    {
        return possibleActions;
    }

    /**
     * Returns whether the current state is terminal.
     * 
     * Throws a RuntimeException if the state was not set.
     * 
     * @return Is the current state terminal?
     */
    public boolean isAtTerminalState()
    {
        if (currentState == null)
        {
            throw new RuntimeException("Uninitialized MountainCarEnvironment.");
        }  
        if (currentState.getCoordinate(0) >= 0.5)
        {
            return true;
        }
        else
        {
            return false;
        }
    }

    /**
     * Set the current state to the given state.
     * 
     * Throws a RuntimeException if the provided state does not have 2 dimension or
     * if it is out of the range [-1.2, 0.5] X [-0.07, 0.07].
     * 
     * @param state
     */
    public void setState(State state)
    {
        currentState = (EuclideanLocation) state;
        // Creates problems because double's are used to approximate real numbers
        /*if (currentState.getDimensionality() != 2)
        {
            throw new RuntimeException("States in MountainCarEnvironment must be 2-dimensional.");
        }
        if (currentState.getCoordinate(0) < -1.2 || 
                currentState.getCoordinate(0) > 0.5 ||
                currentState.getCoordinate(1) < -0.07 ||
                currentState.getCoordinate(1) > 0.07)
        {
            throw new RuntimeException("States in MountainCarEnvironment must be in the interval [-1.2, 0.5] X [-0.07, 0.07].");
        }*/
    }

    /**
     * @return
     */
    public StateSpace getStateSpace()
    {  
        return stateSpace;
    }   

    /**
    *
    */
   class MountainCarTransitionProbabilities implements TransitionProbabilities
   {
       
       /**
        * @param state
        * @param action
        * @return
        */
       public ProbabilityMeasure getTransitionProbabilityMeasure(State state, Action givenAction)
       {
	        EuclideanLocation givenState = (EuclideanLocation) state;
	        double actionValue = ((DiscreteAction) givenAction).getValue();
	        if (actionValue != 1 && actionValue != -1)
	        {
	            throw new RuntimeException("Illegal action selected in MountainCarEnvironment.");
	        }
	        double currentPosition = givenState.getCoordinate(0);
	        double currentVelocity = givenState.getCoordinate(1);       
	        
	        // Compute the new velocity using:
	        // delta velocity = (acceleration of the car - gravity accel.) * delta time
	        double nextVelocity = currentVelocity + 0.001 * actionValue
				- 0.0025 * Math.cos((double) 3 * currentPosition);
	        if (nextVelocity < -0.07) 
	        {
	            nextVelocity = -0.07;
	        }
	        if (nextVelocity > 0.07)
	        {
	            nextVelocity = 0.07;
	        }
	        
	        // Compute the new position using:
	        // delta position = velocity * delta time
	        double nextPosition = currentPosition + nextVelocity;
	        if (nextPosition < -1.2) 
			{
			    nextPosition = -1.2; 
			    nextVelocity = 0.0;
			}
	        if (nextPosition >= 0.5)
	        {
	            nextPosition = 0.5;
	            currentReward = 0.0;
	        }     
			
	        EuclideanLocation nextState = new EuclideanLocation(2);
	        nextState.setCoordinate(0, nextPosition);
	        nextState.setCoordinate(1, nextVelocity);
	        
	        List probabilities = new ArrayList(1);
	        probabilities.add(0, new Double(1));
	        List states = new ArrayList(1);
	        states.add(0, nextState);
	        
	        DiscreteMeasure measure = new DiscreteMeasure(states, probabilities);
	        measure.setStateSpace(stateSpace);
	        return measure;
       }
    }

    /**
     * @return
     */
    public TransitionProbabilities getTransitionProbabilities()
    {
        return new MountainCarTransitionProbabilities();
    }
}

