/**
 * @file 	SARSAAgent.java
 * 
 * @date 	Aug 4, 2004
 * @author 	Alexandre Bouchard
 */

package ca.mcgill.rl.ndp;

import java.io.File;
import java.util.List;

import ca.mcgill.rl.util.AffineVector;
import ca.mcgill.rl.util.Configurable;
import ca.mcgill.rl.util.Gaugeable;
import ca.mcgill.rl.util.RealSequence;
import ca.mcgill.rl.util.State;
import ca.mcgill.rl.util.TypedProperties;

/**
 * An agent using the SARSA algorithm to learn and act in its environment.
 * References:	- @link http://www-anw.cs.umass.edu/~rich/book/the-book.html
 * 				- Implementation of SDM by Bohdana Ratitch
 */
public class SARSAAgent implements Agent, Gaugeable, Configurable
{
    /**
     * The discount factor of the problem this agent is solving.
     */
    private double discountFactor;
    
    /**
     * The state of the last call of actAndLearn.
     */
    private State backedUpState;
    
    /**
     * TODO:
     * Comment for <code>backedUpAction</code>
     */
    private Action backedUpAction;
    
    /**
     * TODO:
     * Comment for <code>learningPolicy</code>
     */
    private ImplicitStateActionPolicy learningPolicy;
    
    /**
     * TODO:
     * Comment for <code>greedyPolicy</code>
     */
    private Policy greedyPolicy;
    
    /**
     * TODO:
     * Comment for <code>stateActionApproximator</code>
     */
    private StateActionApproximator stateActionApproximator;
    
    /**
     * TODO:
     * Comment for <code>stepSizeSequence</code>
     */
    private RealSequence stepSizeSequence;
    
    /**
     * TODO:
     * Comment for <code>iterationNumber</code>
     */
    private int iterationNumber = 0;
    
    /**
     * TODO:
     * Comment for <code>propertyPrefix</code>
     */
    private static String propertyPrefix = "SARSAAgent.";

    /**
     * TODO:
     * Comment for <code>environment</code>
     */
    private EnvironmentSimulator environment;
    
    /*
     * Creates an instance of a SARSA agent. Before using the methods act()
     * and actAndLearn(), the user should call initialize() first.
     * 
     * Note: The policy provided should be derived from the Q-values (e.g., 
     * an SimpleImplicitStateActionPolicy or an EpsilonGreedyStateActionPolicy built from an
     * SimpleImplicitStateActionPolicy).
     * 
     * @param	
     * @param	stateActionApproximator The approximation architecture used to represent Q values
     * @param	implicitPolicy The policy that SARSA uses to choose actions
     
    public SARSAAgent(StateActionApproximator stateActionApproximator, Policy learningPolicy, RealSequence stepsizeSequence)
    {
		this.learningPolicy = learningPolicy;
		this.stateActionApproximator = stateActionApproximator;
		this.stepSizeSequence = stepsizeSequence;
		greedyPolicy = new SimpleImplicitStateActionPolicy(stateActionApproximator);
    }*/
    
    public void setDiscountFactor(double discountFactor)
    {
        this.discountFactor = discountFactor;
    }
    
    /**
     * The current trajectory (if any) is cleared, and the agent will begin the
     * next trajectory at the given state.
     */
    public void initialize()
    {
        backedUpState = null;
        backedUpAction = null;
    }    

    /**
     * Act greedily according to the implicit greedy policy corresponding to the
     * Q values estimated by the approximator
     * 
     * @return The recommended action
     */
    public Action act()
    {
        State currentState = environment.getState();
        List possibleActions = environment.getPossibleActions();
        return greedyPolicy.selectAction(currentState, possibleActions);
    }

    /**
     * Act according to the learningPolicy and update the approximator using the
     * SARSA learning algorithm.
     * 
     * @return The recommended action
     */
    public Action actAndLearn()
    {
        State currentState = environment.getState(); 
        double reward = environment.getReward();
        List possibleActions = environment.getPossibleActions();
        if (backedUpState == null)
        {
            // We need full information on a transition to do learning, so there is 
            // no learning for the first call of actAndLearn
            backedUpAction = learningPolicy.selectAction(currentState, possibleActions);
            backedUpState = currentState;
            return backedUpAction;
        }
        else
        {
            // The SARSA algorithm
            Action currentAction = learningPolicy.selectAction(currentState, possibleActions);
            double temporalDifference = reward
            		+ discountFactor*stateActionApproximator.predict(currentState, currentAction)
            		- stateActionApproximator.predict(backedUpState, backedUpAction);
            AffineVector parametersIncrement = 
                stateActionApproximator.computeGradient(backedUpState, backedUpAction);
            parametersIncrement.times(temporalDifference);
            parametersIncrement.times(stepSizeSequence.getTerm(iterationNumber));
            stateActionApproximator.updateParameters(parametersIncrement, backedUpAction);
            iterationNumber = iterationNumber + 1; 
            backedUpAction = currentAction;
            backedUpState = currentState;
            return currentAction;
        }
    }

    /**
     * 
     */
    public void printStatistics(File directory)
    {
        stateActionApproximator.printStatistics(directory);
    }

    /**
     * @param environment
     */
    public void setEnvironment(EnvironmentSimulator environment)
    {
        this.environment = environment;
    }

    /**
     * @param properties
     */
    public void configure(TypedProperties properties)
    {
        stepSizeSequence = (RealSequence) properties.getPropertyAsObject(propertyPrefix + "stepSizeSequence");
        if (stepSizeSequence instanceof Configurable)
        {
            ((Configurable) stepSizeSequence).configure(properties);
        }
        
        stateActionApproximator = new StateActionApproximator();
        stateActionApproximator.setEnvironment(environment);
        stateActionApproximator.configure(properties);
        
        learningPolicy = (ImplicitStateActionPolicy) properties.getPropertyAsObject(propertyPrefix + "learningPolicy");
        learningPolicy.setStateActionApproximator(stateActionApproximator);
        if (learningPolicy instanceof Configurable)
        {
            ((Configurable) learningPolicy). configure(properties);
        }
        greedyPolicy = new SimpleImplicitStateActionPolicy(stateActionApproximator);
    }
}
