/**
 * @file 	SimpleImplicitStateActionPolicy.java
 * 
 * @date 	Aug 4, 2004
 * @author 	Alexandre Bouchard
 */

package ca.mcgill.rl.ndp;

import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;
import java.util.Random;

import ca.mcgill.rl.util.State;

/**
 * A greedy policy constructed from a StateAction approximator.
 */
public class SimpleImplicitStateActionPolicy implements ImplicitStateActionPolicy
{
    /**
     * A time-seeded pseudo-random generator.
     */
    private Random randomNumberGenerator = new Random(System.currentTimeMillis());
    
    /**
     * The underlying Q values approximator.
     */
    private StateActionApproximator approximator;
    
    /**
     * Creates an implicit, greedy policy from the supplied approximator.
     * 
     * Throws a RuntimeException if the provided approximator is null.
     * 
     * @param approximator
     */
    public SimpleImplicitStateActionPolicy(StateActionApproximator approximator)
    {
        if (approximator == null)
        {
           throw new RuntimeException("SimpleImplicitStateActionPolicy must be initialized with a non null approximator"); 
        }
        this.approximator = approximator;
    }
    
    /**
     * 
     */
    public SimpleImplicitStateActionPolicy()
    {
    }
    
    /**
     * Selects greedily an action.
     * 
     * The subset of the possibleActions that are maximal according to the StateAction approximator
     * is computed first. Then one action is picked randomly in this subset.
     * 
     * @param currentState The state from which the decision is taken
     * @param possibleActions The actions that are allowed from the currentState
     * @return The greedily selected action
     */
    public Action selectAction(State currentState, List possibleActions)
    {
        if (possibleActions.isEmpty())
        {
            throw new RuntimeException("No action available at current state " + currentState.toString());
        }
        // 1- Find the subset of possibleActions with the highest value according to the approximator
        List optimalActions = new ArrayList();
        Iterator possibleActionsIterator = possibleActions.iterator();
        double maxValue = -Double.MAX_VALUE;
        while (possibleActionsIterator.hasNext())
        {
            Action currentAction = (Action) possibleActionsIterator.next();
            double currentValue = approximator.predict(currentState, currentAction);
            if (currentValue > maxValue)
            {
                optimalActions.clear();
                optimalActions.add(currentAction);
                maxValue = currentValue;
            }
            if (currentValue == maxValue)
            {
                optimalActions.add(currentAction);
            }
        }
        // 2- Pick an action from this subset randomly (uniform distribution)
        return (Action) optimalActions.get(randomNumberGenerator.nextInt(optimalActions.size()));
    }

    /**
     * @param stateActionApproximator
     */
    public void setStateActionApproximator(StateActionApproximator stateActionApproximator)
    {
        this.approximator = stateActionApproximator;
    }
}
