/**
 * @file 	StateActionApproximator.java
 * 
 * @date 	Aug 7, 2004
 * @author 	Alexandre Bouchard
 */

package ca.mcgill.rl.ndp;

import java.io.File;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;

import ca.mcgill.rl.util.AffineVector;
import ca.mcgill.rl.util.Configurable;
import ca.mcgill.rl.util.Gaugeable;
import ca.mcgill.rl.util.State;
import ca.mcgill.rl.util.TypedProperties;

/**
 * A state-action (Q values) approximator.
 * 
 * Implemented by wrapping a map action-approximator and by dispatching to the
 * appropriate approximator in this map.
 */
public class StateActionApproximator implements Gaugeable, Configurable
{
    /**
     * The underlying action-approximator map.
     */
    private Map approximators;
    
    /**
     * TODO:
     * Comment for <code>environment</code>
     */
    private EnvironmentSimulator environment;
    
    /**
     * TODO:
     * Comment for <code>propertyPrefix</code>
     */
    private static String propertyPrefix = "StateActionApproximator.";
    
    /*
     * Creates a state-action approximator around the given map of approximators.
     * 
     * Thow a RuntimeException if the given Map is null.
     * 
     * @param 	approximators
     
    public StateActionApproximator(Map approximators)
    {
        if (approximators == null)
        {
            throw new RuntimeException("StateActionApproximator must be initialized with a nonnull approximator.");
        }
        this.approximators = approximators;
    }*/
    
    /**
     * 
     */
    public StateActionApproximator()
    {
    }
    
	/**
	 * Increments the parameters by a given amount for the approximator 
	 * corresponding to the given action.
	 * 
	 * The action must be a key in the map approximators.
	 * @param	parametersIncrement The amount the parameters should be incremented
	 * 			in each dimension.
	 * @param multiplicativeConstant TODO
	 * @param	action The given action.
	 */
	public void updateParameters(AffineVector parametersIncrement, Action action)
	{
	    ((Approximator) approximators.get(action)).updateParameters(parametersIncrement);
	}
	
	/**
	 * Evaluates the approximation function at the given point in the state space
	 * for the approximator corresponding to the given action.
	 * 
	 * The action must be a key in the map approximators.
	 * 
	 * @param	action The given action
	 * @param	state The point in the state space where the function should be
	 * 			evaluated.
	 */
	public double predict(State state, Action action)
	{
	    return ((Approximator) approximators.get(action)).evaluateAt(state);
	}
	
	/**
	 * Evaluates the gradient of the approximation function at the given point in the
	 * state space for the approximator corresponding to the given action.
	 * 
	 * The action must be a key in the map approximators.
	 * 
	 * @param	action The given action
	 * @param	state The point in the state space where the gradient should be
	 * 			evaluated
	 */
	public AffineVector computeGradient(State state, Action action)
	{
	    return ((Approximator) approximators.get(action)).computeGradient(state);
	}

    /**
     * 
     */
    public void printStatistics(File directory)
    {
        Iterator approximatorsIterator = approximators.values().iterator();
        int i = 0;
        while (approximatorsIterator.hasNext())
        {
            Object currentObject = approximatorsIterator.next();
            if (currentObject instanceof Gaugeable)
            {
                File subDirectory = new File(directory, "approximator_" + i);
                i = i + 1;
                if (!subDirectory.isDirectory())
                {
                    subDirectory.mkdir();
                }
                ((Gaugeable) currentObject).printStatistics(subDirectory);
            }
        }     
    }

    /**
     * @param properties
     */
    public void configure(TypedProperties properties)
    {
        approximators = new HashMap();
        Iterator actionsIterator = environment.getPossibleActions().iterator();
        while (actionsIterator.hasNext())
        {
            Object currentAction = actionsIterator.next();
            Approximator currentApproximator = (Approximator) properties.getPropertyAsObject(propertyPrefix + "approximators");
            currentApproximator.setStateSpace(environment.getStateSpace());
            if (currentApproximator instanceof Configurable)
            {
                ((Configurable) currentApproximator).configure(properties);
            }
            approximators.put(currentAction, currentApproximator);
        }
    }

    /**
     * @param environment
     */
    public void setEnvironment(EnvironmentSimulator environment)
    {
        this.environment = environment;
    }
}
