/**
 * @file 	TDAgent.java
 * 
 * @date 	Aug 17, 2004
 * @author 	Alexandre Bouchard
 */

package ca.mcgill.rl.ndp;

import java.io.File;
import java.util.List;

import ca.mcgill.rl.util.AffineVector;
import ca.mcgill.rl.util.Configurable;
import ca.mcgill.rl.util.Gaugeable;
import ca.mcgill.rl.util.RealSequence;
import ca.mcgill.rl.util.State;
import ca.mcgill.rl.util.TypedProperties;

/**
 *
 */
public class TDAgent implements Agent, Configurable, Gaugeable
{
    /**
     * The state of the last call of actAndLearn.
     */
    private State backedUpState;
    
    /**
     * TODO:
     * Comment for <code>propertyPrefix</code>
     */
    private static String propertyPrefix = "TDAgent.";
    
    /**
     * TODO:
     * Comment for <code>learningPolicy</code>
     */
    private Policy learningPolicy;
    
    /**
     * TODO:
     * Comment for <code>greedyPolicy</code>
     */
    private SimpleImplicitCostToGoPolicy greedyPolicy;
    
    /**
     * TODO:
     * Comment for <code>approximator</code>
     */
    private Approximator approximator;
    
    /**
     * TODO:
     * Comment for <code>discountFactor</code>
     */
    private double discountFactor;
    
    /**
     * TODO:
     * Comment for <code>iterationNumber</code>
     */
    private int iterationNumber = 0;    
    
    /**
     * TODO:
     * Comment for <code>stepSizeSequence</code>
     */
    private RealSequence stepSizeSequence;    
    
    /**
     * TODO:
     * Comment for <code>environment</code>
     */
    private ModelledEnvironmentSimulator environment;

    /**
     * 
     */
    public void initialize()
    {
        backedUpState = null;
    }

    /**
     * @return
     */
    public Action act()
    {
        State currentState = environment.getState();
        List possibleActions = environment.getPossibleActions();
        return greedyPolicy.selectAction(currentState, possibleActions);
    }

    /**
     * @return
     */
    public Action actAndLearn()
    {
        State currentState = environment.getState(); 
        double reward = environment.getReward();
        List possibleActions = environment.getPossibleActions();
        if (backedUpState == null)
        {
            // We need full information on a transition to do learning, so there is 
            // no learning for the first call of actAndLearn
            backedUpState = currentState;
            return learningPolicy.selectAction(currentState, possibleActions);
        }
        else
        {        
            // The TD algorithm
            double temporalDifference = reward
            		+ discountFactor*approximator.evaluateAt(currentState)
            		- approximator.evaluateAt(backedUpState);
            AffineVector parametersIncrement = 
                approximator.computeGradient(backedUpState);
            parametersIncrement.times(temporalDifference);
            parametersIncrement.times(stepSizeSequence.getTerm(iterationNumber));
            approximator.updateParameters(parametersIncrement);
            iterationNumber = iterationNumber + 1; 
            backedUpState = currentState;
            return learningPolicy.selectAction(currentState, possibleActions);            
        }
    }

    /**
     * @param discountFactor
     */
    public void setDiscountFactor(double discountFactor)
    {
        this.discountFactor = discountFactor;
    }

    /**
     * @param environment
     */
    public void setEnvironment(EnvironmentSimulator environment)
    {
        this.environment = (ModelledEnvironmentSimulator) environment;
    }

    /**
     * @param properties
     */
    public void configure(TypedProperties properties)
    {
        stepSizeSequence = (RealSequence) properties.getPropertyAsObject(propertyPrefix + "stepSizeSequence");
        if (stepSizeSequence instanceof Configurable)
        {
            ((Configurable) stepSizeSequence).configure(properties);
        }
        
        approximator = (Approximator) properties.getPropertyAsObject(propertyPrefix + "approximator");
        approximator.setStateSpace(environment.getStateSpace());
        if (approximator instanceof Configurable)
        {
            ((Configurable) approximator).configure(properties);
        }
        
        learningPolicy = (ImplicitCostToGoPolicy) properties.getPropertyAsObject(propertyPrefix + "learningPolicy");
        ((ImplicitCostToGoPolicy) learningPolicy).setApproximator(approximator);
        ((ImplicitCostToGoPolicy) learningPolicy).setTransitionProbabilities(environment.getTransitionProbabilities());
        if (learningPolicy instanceof Configurable)
        {
            ((Configurable) learningPolicy). configure(properties);
        }
        greedyPolicy = new SimpleImplicitCostToGoPolicy();
        greedyPolicy.setApproximator(approximator);
        greedyPolicy.setTransitionProbabilities(environment.getTransitionProbabilities());
    }

    /**
     * @param experimentDirectory
     */
    public void printStatistics(File experimentDirectory)
    {
        if (approximator instanceof Gaugeable)
        {
	        File subDirectory = new File(experimentDirectory, "approximator");
	        if (!subDirectory.isDirectory())
	        {
	            subDirectory.mkdir();
	        }
	        ((Gaugeable) approximator).printStatistics(subDirectory);    
	    }
    }
}
