/**
 * @file 	EuclideanSDMArchitecture.java
 * 
 * @date 	Aug 4, 2004
 * @author 	Alexandre Bouchard
 */

package ca.mcgill.rl.sdm;

import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.PrintWriter;
import java.util.Iterator;

import ca.mcgill.rl.ndp.Approximator;
import ca.mcgill.rl.util.AffineVector;
import ca.mcgill.rl.util.Configurable;
import ca.mcgill.rl.util.EuclideanLocation;
import ca.mcgill.rl.util.Gaugeable;
import ca.mcgill.rl.util.MultiInterval;
import ca.mcgill.rl.util.StandardVector;
import ca.mcgill.rl.util.State;
import ca.mcgill.rl.util.StateSpace;
import ca.mcgill.rl.util.TypedProperties;

/**
 * A Sparse Distributed Memory architecture for euclidean spaces.
 * 
 * A linear, local approximator with adaptative basis points (called hard locations).
 */
public class EuclideanSDMArchitecture implements Approximator, Gaugeable, Configurable
{
    /**
     * TODO:
     * Comment for <code>propertyPrefix</code>
     */
    private static String propertyPrefix = "EuclideanSDMArchitecture.";
    
    /**
     * The special data structure to store and retreive the hard locations.
     */
    private HardLocationSupport hardLocationSupport;
    
    /**
     * The system managing the dynamic insertion/deletion of hard locations.
     */
    private HardLocationManager hardLocationManager;

    /**
     * TODO:
     * Comment for <code>stateSpace</code>
     */
    private MultiInterval stateSpace;

    /**
     * Updates the parameters of this architecture (i.e. the weights of the hard
     * locations)
     * @param parametersIncrement
     */
    public void updateParameters(AffineVector parametersIncrement)
    {
        hardLocationSupport.updateParameters(parametersIncrement);
    }

    /**
     * Evalutates the approximation function at the given state.
     * 
     * Using the local property of the approximator, only the hard locations in a
     * neighborhood around the state affect the evaluation. This neighborhood is 
     * first computed by hardLocationSupport.getNeighborhood(). Then the evaluation
     * is done by neighborhood.inducedWeightOnCenter().
     * 
     * The dynamic basis point reorganization method hardLocationManager.reorganize()
     * is also called, after that the evaluation is computed.
     * 
     * @param state
     * @return The evaluation.
     */
    public double evaluateAt(State state)
    {   
        EuclideanLocation location = (EuclideanLocation) state;     
        ActiveLocations neighborhood = hardLocationSupport.getNeighborhood(location);    
        double returnedValue = inducedWeightOnCenter(neighborhood); 
        hardLocationManager.reorganize(neighborhood);
        return returnedValue;
    }
    
    /**
     * Returns the weight induced on the center or 0 if there are no hard locations
     * activated.
     * 
     * It uses the formula:
     * sum_{hard locations h_k} mu(center, h_k) weight(h_k) /
     * 		sum{hard locations h_k} mu(center, h_k), where mu(.,.) is the similarity
     * metric of the corresponding hard location.
     * 
     * @return The induced weight.
     */    
    public double inducedWeightOnCenter(ActiveLocations neighborhood)
    {
        double denominatorPartialSum = 0;
        double partialSum = 0;
        Iterator iterator = neighborhood.getHardLocations().iterator();
        while (iterator.hasNext())
        {
            HardLocation currentHardLocation = (HardLocation) iterator.next();
            double distanceFromCenter = currentHardLocation.similarityWith(neighborhood.getCenter());
            denominatorPartialSum = denominatorPartialSum + distanceFromCenter;
            partialSum = partialSum + distanceFromCenter * hardLocationSupport.getWeight(currentHardLocation);
        }
        if (denominatorPartialSum == 0)
        {
            return 0;
        }
        else
        {
            return partialSum/denominatorPartialSum;
        }    
    }

    /**
     * Computes the gradient of the approximation function at the given state.
     * 
     * The algorithm also used the fact that only the hard locations in a
     * neighborhood centered at the given state affect the gradient.
     * 
     * The gradient of the m-th component is therefore 0 if the corresponding hard
     * location is outside the neighborhood, otherwise it is:
     * mu(h_m, state) / sum_{hard locations h_k} mu(h_k, state), where mu is the
     * similarity metric (defined in the hard location implementation chosen by the
     * hard location manager).
     * 
     * @param state
     * @return The gradient vector.
     */
    public AffineVector computeGradient(State state)
    {
        EuclideanLocation location = (EuclideanLocation) state;    
        AffineVector returnedVector = new StandardVector(hardLocationSupport.getNumberOfParameters());
        ActiveLocations neighborhood = hardLocationSupport.getNeighborhood(location);
        
        double denominatorPartialSum = 0;
        Iterator iterator = neighborhood.getHardLocations().iterator();
        while (iterator.hasNext())
        {
            HardLocation currentHardLocation = (HardLocation) iterator.next();
            double distanceFromCenter = currentHardLocation.similarityWith(location);
            denominatorPartialSum = denominatorPartialSum + distanceFromCenter;
            returnedVector.setComponent(currentHardLocation.getId(), distanceFromCenter);
        }
        
        if (denominatorPartialSum == 0)
        {
            return returnedVector;
        }
        else
        {
            returnedVector.times((double) ((double) 1/(double) denominatorPartialSum));
            return returnedVector;
        }
    }
    
    /**
     * TODO:
     * Comment for <code>statisticsIterator</code>
     */
    private int statisticsIterator = 0;
    
    /**
     * TODO:
     * Comment for <code>numberOfGrids</code>
     */
    private int numberOfGrids;

    /**
     * 
     */
    public void printStatistics(File directory)
    {
        if (hardLocationSupport instanceof Gaugeable)
        {
            File subDirectory = new File(directory, "hardLocationSupport");
            if (!subDirectory.isDirectory())
            {
                subDirectory.mkdir();
            }
            ((Gaugeable) hardLocationSupport).printStatistics(subDirectory);
        } 
        if (hardLocationManager instanceof Gaugeable)
        {
            File subDirectory = new File(directory, "hardLocationManager");
            if (!subDirectory.isDirectory())
            {
                subDirectory.mkdir();
            }            
            ((Gaugeable) hardLocationManager).printStatistics(subDirectory);      
        }
        
        try
        {
            String postFix = ".ndd";
            if (stateSpace.getDimensionality() == 2)
            {
                postFix = ".3dd";
            }
            if (stateSpace.getDimensionality() == 1)
            {
                postFix = ".2dd";
            }   
            PrintWriter out = new PrintWriter(new FileOutputStream(new File(directory, "" + statisticsIterator + postFix)));
            statisticsIterator = statisticsIterator + 1;
            EuclideanLocation location = new EuclideanLocation(stateSpace.getDimensionality());
            for (int i = 0; i < stateSpace.getDimensionality(); i++)
            {
                location.setCoordinate(i, stateSpace.getLowerBound(i));
            }
            recursivePrintStatistics(out, location, stateSpace.getDimensionality() - 1);
            out.close();
        }
        catch (FileNotFoundException e)
        {
            // TODO Auto-generated catch block
            e.printStackTrace();
        }         
    }

    /**
     * @param out
     * @param i
     */
    private void recursivePrintStatistics(PrintWriter out, EuclideanLocation currentLocation, int currentDimension)
    {
        if (currentDimension == 0)
        {
            double increment = (stateSpace.getUpperBound(0) - stateSpace.getLowerBound(0)) / numberOfGrids;
            currentLocation.setCoordinate(0, stateSpace.getLowerBound(0));
            for (int i = 0; i <= numberOfGrids; i++)
            {
                ActiveLocations neighborhood = hardLocationSupport.getNeighborhood(currentLocation);    
                double evaluation = inducedWeightOnCenter(neighborhood); 
                for (int j = 0; j < stateSpace.getDimensionality(); j++)
                {
                    out.print("" + currentLocation.getCoordinate(j) + " ");
                }
                out.print(evaluation);
                out.println();
                currentLocation.setCoordinate(0, currentLocation.getCoordinate(0) + increment);
            }
        }
        else
        {
            double increment = (stateSpace.getUpperBound(currentDimension) - stateSpace.getLowerBound(currentDimension)) / numberOfGrids;
            currentLocation.setCoordinate(currentDimension, stateSpace.getLowerBound(currentDimension));
            for (int i = 0; i <= numberOfGrids; i++)
            {
                recursivePrintStatistics(out, currentLocation, currentDimension - 1);
                currentLocation.setCoordinate(currentDimension, currentLocation.getCoordinate(currentDimension) + increment);
            }
        }
    }

    /**
     * @param stateSpace
     */
    public void setStateSpace(StateSpace stateSpace)
    {
        this.stateSpace = (MultiInterval) stateSpace;
    }

    /**
     * @param properties
     */
    public void configure(TypedProperties properties)
    {
        numberOfGrids = properties.getPropertyAsInt(propertyPrefix + "numberOfGrids");
        hardLocationManager = (HardLocationManager) properties.getPropertyAsObject(propertyPrefix + "hardLocationManager");
        hardLocationSupport = (HardLocationSupport) properties.getPropertyAsObject(propertyPrefix + "hardLocationSupport");
        hardLocationSupport.setSDM(this);
        hardLocationManager.setSDM(this);
  
        if (hardLocationManager instanceof Configurable)
        {
            ((Configurable) hardLocationManager).configure(properties);
        }        
        if (hardLocationSupport instanceof Configurable)
        {
            ((Configurable) hardLocationSupport).configure(properties);
        }
        hardLocationManager.initialInsertion();
    }

    /**
     * @return
     */
    public StateSpace getStateSpace()
    {
        return stateSpace;
    }

    /**
     * @return
     */
    public HardLocationSupport getHardLocationSupport()
    {
        return hardLocationSupport;
    }
    
    /**
     * @return
     */
    public HardLocationManager getHardLocationManager()
    {
        return hardLocationManager;
    }
}
