/**
 * @file 	HashSupport.java
 * 
 * @date 	Aug 9, 2004
 * @author 	Alexandre Bouchard
 */

package ca.mcgill.rl.sdm;

import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.PrintWriter;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;

import ca.mcgill.rl.util.AffineVector;
import ca.mcgill.rl.util.Configurable;
import ca.mcgill.rl.util.EuclideanLocation;
import ca.mcgill.rl.util.Gaugeable;
import ca.mcgill.rl.util.MultiInterval;
import ca.mcgill.rl.util.SparseVector;
import ca.mcgill.rl.util.TypedProperties;

/**
 * An implementation of HardLocationSupport based on a hash.
 * 
 * It assumes that the hard location have the same activation radius and partition
 * the space into hyper rectangular cells of the same edge size as the activation
 * radius. The partition index is used as a key in the hash. The hash map is
 * actually a multi map because it is Lists of Hardlocations that are stored.
 */
public class HashSupport implements HardLocationSupport, Gaugeable, Configurable
{
    /**
     * The hard locations, indexed by id.
     */
    private AffineVector weights;
    
    /**
     * The hard location, indexed by position.
     */
    private Map indexByPosition = new HashMap();    
    
    /**
     * The activation radius.
     */
    private List radius;

    /**
     * TODO:
     * Comment for <code>sdm</code>
     */
    private EuclideanSDMArchitecture sdm;
    
    /**
     * 
     */
    public HashSupport()
    {
    }
    
    /**
     * @param parametersIncrement
     */
    public void updateParameters(AffineVector parametersIncrement)
    {
        weights.add(parametersIncrement);
    }

    /**
     * @param state
     * @return
     */
    public ActiveLocations getNeighborhood(EuclideanLocation state)
    {
        List listOfNeighborCellsRepresentatives = getNeighborCellsRepresentatives(state);
        List neighborHardLocations = new ArrayList();
        Iterator iterator = listOfNeighborCellsRepresentatives.iterator();
        while (iterator.hasNext())
        {
            EuclideanLocation currentRepresentative = (EuclideanLocation) iterator.next();
            List currentList = (List) indexByPosition.get(getKey(currentRepresentative));
            if (currentList != null)
            {
                Iterator iterator2 = currentList.iterator();
                while (iterator2.hasNext())
                {
                    HardLocation currentHardLocation = (HardLocation) iterator2.next();
                    if (currentHardLocation.similarityWith(state) > 0)
                    {
                        neighborHardLocations.add(currentHardLocation);
                    }
                }
            }
        }
        return new ActiveLocations(state, neighborHardLocations);
    }

    /**
     * @param state
     * @return
     */
    private List getNeighborCellsRepresentatives(EuclideanLocation state)
    {
        return getNeighborCellsRepresentatives(state, ((MultiInterval) sdm.getStateSpace()).getDimensionality() - 1);
    }

    /**
     * @param state
     * @param currentDimension
     * @return
     */
    private List getNeighborCellsRepresentatives(EuclideanLocation state, int currentDimension)
    {
        List returnedList;
        if (currentDimension == 0)
        {
            double currentRadius = ((Double) radius.get(0)).doubleValue();
            returnedList = new ArrayList();
            EuclideanLocation state1 = new EuclideanLocation(state);
            state1.setCoordinate(0, state.getCoordinate(0) + currentRadius);
            returnedList.add(state1);
            if ((state.getCoordinate(0) - currentRadius) > ((MultiInterval) sdm.getStateSpace()).getLowerBound(0))
            {
                EuclideanLocation state2 = new EuclideanLocation(state);
                state2.setCoordinate(0, state.getCoordinate(0) - currentRadius);
                returnedList.add(state2);
            }
        }
        else
        {
            double currentRadius = ((Double) radius.get(currentDimension)).doubleValue(); 
            EuclideanLocation state1 = new EuclideanLocation(state);
            state1.setCoordinate(currentDimension, state.getCoordinate(currentDimension) + currentRadius);
            returnedList = getNeighborCellsRepresentatives(state1, currentDimension - 1);
            if ((state.getCoordinate(currentDimension) - currentRadius) > ((MultiInterval) sdm.getStateSpace()).getLowerBound(0))
            {
                EuclideanLocation state2 = new EuclideanLocation(state);
                state2.setCoordinate(currentDimension, state.getCoordinate(currentDimension) - currentRadius);
                returnedList.addAll(getNeighborCellsRepresentatives(state2, currentDimension - 1));
            }
        }
        return returnedList;
    }

    /**
     * TODO: could be problematic... if there is deletion...
     * 
     * @return
     */
    public int getNumberOfParameters()
    {
        if (weights == null)
        {
            return 0;
        }
        else
        {
            return weights.getDimensionality();
        }
    }

    /**
     * @param newHardLocation
     */
    public void addHardLocation(HardLocation newHardLocation)
    {
        if (weights == null)
        {
            weights = new SparseVector(1);
        }
        else
        {
            weights.setDimensionality(weights.getDimensionality() + 1);
        }
        
        List currentlyStored = (List) indexByPosition.get(getKey(newHardLocation.getPosition()));
        if (currentlyStored == null)
        {
            currentlyStored = new ArrayList();
            currentlyStored.add(newHardLocation);
            indexByPosition.put(getKey(newHardLocation.getPosition()), currentlyStored);
        }
        else
        {
            currentlyStored.add(newHardLocation);
        }
    }

    /**
     * @param hardLocation
     */
    public void removeHardLocation(HardLocation hardLocation)
    {
        setWeight(hardLocation, 0);
        List currentList = (List) indexByPosition.get(getKey(hardLocation.getPosition()));
        for (int i = 0; i < currentList.size(); i++)
        {
            HardLocation currentHardLocation = (HardLocation) currentList.get(i);
            if (currentHardLocation == hardLocation)
            {
                currentList.remove(i);
            }
        }
        if (currentList.size() == 0)
        {
            indexByPosition.remove(getKey(hardLocation.getPosition()));
        }
    }
    
    protected String getKey(EuclideanLocation location)
    {
        StringBuffer buffer = new StringBuffer();
        for (int i = 0; i < location.getDimensionality(); i++)
        {
            int currentPartitionIndex = 
                (int) ((location.getCoordinate(i) - ((MultiInterval) sdm.getStateSpace()).getLowerBound(i)) 
                    / ((double) 2 * ((Double) radius.get(i)).doubleValue()));
            buffer.append((new Integer(currentPartitionIndex)).toString());
            buffer.append(",");
        }
        return buffer.toString();
    }
    
    /**
     * TODO:
     * Comment for <code>statisticsIterator</code>
     */
    private int statisticsIterator = 0;

    /**
     * 
     */
    public void printStatistics(File directory)
    {   
        try
        {
            String postFix = ".ndd";
            if (((MultiInterval) sdm.getStateSpace()).getDimensionality() == 2)
            {
                postFix = ".2dd";
            }           
            PrintWriter out = new PrintWriter(new FileOutputStream(new File(directory, "" + statisticsIterator + postFix)));
            statisticsIterator = statisticsIterator + 1;
            Iterator hardLocationsIterator = indexByPosition.values().iterator();
            while (hardLocationsIterator.hasNext())
            {
                List currentHardLocations = (List) hardLocationsIterator.next();
                Iterator hardLocationIterator = currentHardLocations.iterator();
                while (hardLocationIterator.hasNext())
                {
                    HardLocation currentHardLocation = (HardLocation) hardLocationIterator.next();
	                for (int i = 0; i < currentHardLocation.getPosition().getDimensionality(); i++)
	                {
	                    out.print("" + currentHardLocation.getPosition().getCoordinate(i) + " ");
	                }
	                out.println();
                }
            }
            out.close();
        }
        catch (FileNotFoundException e)
        {
            // TODO Auto-generated catch block
            e.printStackTrace();
        }             
    }

    /**
     * @param hardLocation
     * @return
     */
    public double getWeight(HardLocation hardLocation)
    {
        return weights.getComponent(hardLocation.getId());
    }

    /**
     * @param hardLocation
     * @param weight
     */
    public void setWeight(HardLocation hardLocation, double weight)
    {
        weights.setComponent(hardLocation.getId(), weight);
    }

    /**
     * @param architecture
     */
    public void setSDM(EuclideanSDMArchitecture architecture)
    {
        this.sdm = architecture;
    }

    /**
     * @param properties
     */
    public void configure(TypedProperties properties)
    {
        radius = ((ConstantRadiusHardLocationManager) sdm.getHardLocationManager()).getRadius();
    }

}
