/**
 * @file 	SimpleManager.java
 * 
 * @date 	Aug 9, 2004
 * @author 	Alexandre Bouchard
 */

package ca.mcgill.rl.sdm;

import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;
import java.util.Random;

import ca.mcgill.rl.util.Configurable;
import ca.mcgill.rl.util.EuclideanLocation;
import ca.mcgill.rl.util.MultiInterval;
import ca.mcgill.rl.util.TypedProperties;

/**
 * A very simple hard location manager.
 * 
 * Inserts constant radius WeightedMaxNormLocations.
 */
public class SimpleManager implements ConstantRadiusHardLocationManager, Configurable
{
    /**
     * TODO:
     * Comment for <code>propertyPrefix</code>
     */
    private static String propertyPrefix = "SimpleManager.";
    
    /**
     * TODO:
     * Comment for <code>sdm</code>
     */
    private EuclideanSDMArchitecture sdm;
    
    /**
     * The minimum number of hard locations wanted in any neighborhood.
     */
    private int minDensity;
    
    /**
     * TODO:
     * Comment for <code>maxSimilarity</code>
     */
    private double maxSimilarity;
    
    /**
     * The radius for the inserted hard locations.
     */
    private List radius;
    
    /**
     * A time-seeded pseudo-random number generator.
     */
    private Random randomNumberGenerator = new Random(System.currentTimeMillis());
    
    /**
     * The id of the next hard location that will be inserted.
     */
    private int nextId = 0;

    /**
     * TODO:
     * Comment for <code>maxCapacity</code>
     */
    private int maxCapacity;
    
    /*
     * @param radius
     *
    public SimpleManager(List radius)
    {
        this.radius = radius;
    }*/
    
    /**
     * 
     */
    public SimpleManager()
    {
    }

    /**
     * Reorganize the given neighborhood.
     * 
     * Calls randomLocalInsert until the number of hard locations in the neighborhood
     * is at least equal to the field minDensity.
     * 
     * @param neighborhood
     */
    public void reorganize(ActiveLocations neighborhood)
    {
        if (sdm.getHardLocationSupport().getNumberOfParameters() < maxCapacity)
        {
	        while (neighborhood.getHardLocations().size() < minDensity)
	        {
	            randomLocalInsert(neighborhood);
	            neighborhood = sdm.getHardLocationSupport().getNeighborhood(neighborhood.getCenter());
	        }
        }
    }

    /**
     * Pick a position at random in the neighborhood, create a hard location at that
     * position and insert it.
     * 
     * Also set it weight to the value predicted at that random position.
     * 
     * @param neighborhood
     */
    protected void randomLocalInsert(ActiveLocations neighborhood)
    {
        EuclideanLocation randomPosition = null;
        boolean goodPosition = false;
        
        int trials = 0;
        while (!goodPosition && trials < 3)
        {
            trials = trials + 1;
            randomPosition = new EuclideanLocation(((MultiInterval) sdm.getStateSpace()).getDimensionality());
			
			for (int i = 0; i < randomPosition.getDimensionality(); i++)
			{
			    double randomCoordinate = randomNumberGenerator.nextDouble() * 2 * ((Double) radius.get(i)).doubleValue()
			    		+ (neighborhood.getCenter().getCoordinate(i) - ((Double) radius.get(i)).doubleValue());
			    randomPosition.setCoordinate(i, randomCoordinate);
			}
			
			ActiveLocations neighborhoodOfInsertedLocation = sdm.getHardLocationSupport().getNeighborhood(randomPosition);
			Iterator neighborhoodIterator = neighborhoodOfInsertedLocation.getHardLocations().iterator();
			goodPosition = true;
			
			while (neighborhoodIterator.hasNext())
			{
			    HardLocation currentActiveLocation = (HardLocation) neighborhoodIterator.next();
			    if (currentActiveLocation.similarityWith(randomPosition) > maxSimilarity)
			    {
			        goodPosition = false;
			    }
			}
        }
        
        if (goodPosition)
        {
	        WeightedMaxNormHardLocation newHardLocation = new WeightedMaxNormHardLocation(radius, randomPosition, getNextId());
	        double predictedWeight = sdm.inducedWeightOnCenter(neighborhood);
	        newHardLocation.setWeigth(predictedWeight);
	        sdm.getHardLocationSupport().addHardLocation(newHardLocation);
        }
    }

    /**
     * Returns a unique identifier for hard locations.
     * 
     * @return
     */
    private int getNextId()
    {
        int returnedValue = nextId;
        nextId = nextId + 1;
        return returnedValue;
    }

    /**
     * Does nothing; no initial insertion with SimpleManager.
     */
    public void initialInsertion()
    {      
    }

    /**
     * @param architecture
     */
    public void setSDM(EuclideanSDMArchitecture architecture)
    {
        this.sdm = architecture;
    }

    /**
     * @param properties
     */
    public void configure(TypedProperties properties)
    {
        minDensity = properties.getPropertyAsInt(propertyPrefix + "minDensity");
        maxCapacity = properties.getPropertyAsInt(propertyPrefix + "maxCapacity");
        maxSimilarity = properties.getPropertyAsDouble(propertyPrefix + "maxSimilarity", 1);
        radius = new ArrayList(((MultiInterval) sdm.getStateSpace()).getDimensionality());
        for (int i = 0; i < ((MultiInterval) sdm.getStateSpace()).getDimensionality(); i++)
        {
            radius.add(i, new Double(properties.getPropertyAsDouble(propertyPrefix + "radius_" + i)));
        }
    }

    /**
     * @return
     */
    public List getRadius()
    {
        return radius;
    }

}
