/**
 * @file 	WeightedMaxNormHardLocation.java
 * 
 * @date 	Aug 9, 2004
 * @author 	Alexandre Bouchard
 */

package ca.mcgill.rl.sdm;

import java.util.List;

import ca.mcgill.rl.util.EuclideanLocation;

/**
 * An implementation of hard locations that uses a similarity metric that induces
 * weighted max norm-like neighborhoods.
 */
public class WeightedMaxNormHardLocation implements HardLocation
{
    /**
     * The unique identifier of this hard location.
     */
    private int id;
    
    /**
     * The position of this hard location.
     */
    private EuclideanLocation position;
    
    /**
     * The radius, i.e. the "weights" of the similarity metric in each dimension. 
     */
    private List radius;
    
    /**
     * The weight stored at this hard location.
     */
    private double weight;
    
    /**
     * Constructs a new hard location given the radius, center and id. 
     * 
     * The radius is the list of "weights" in each dimension of the similarity 
     * measure (not related to the weight stored in the hard location). They are 
     * called radius because the neighborhood activated by the hard location have 
     * radius radius[i] in the i-th dimension.
     * 
     * Throws a RuntimeException if the radius or the position are null.
     * 
     * @param radius
     * @param position
     * @param id
     */
    public WeightedMaxNormHardLocation(List radius, EuclideanLocation position, int id)
    {
        if (radius == null || position == null)
        {
            throw new RuntimeException("WeightedMaxNormHardLocation must be initialized with nonnull radius and position.");
        }
        this.id = id;
        this.position = position;
        this.radius = radius;
    }

    /**
     * Returns the position in the space of this hard location.
     * 
     * @return
     */
    public EuclideanLocation getPosition()
    {
        // TODO Auto-generated method stub
        return position;
    }

    /**
     * Returns the weight stored in this hard location.
     * 
     * @return
     */
    public double getWeight()
    {
        return weight;
    }
    
    /**
     * Set the weight stored in this hard location.
     * 
     * @param weight
     */
    public void setWeigth(double weight)
    {
        this.weight = weight;
    }    

    /**
     * Uses symmetric triangular functions to compute the similarity with the given
     * location.
     * 
     * Throws a RuntimeException if the location does not have the same
     * dimensionality as the current hard location.
     * 
     * @param location
     * @return
     */
    public double similarityWith(EuclideanLocation location)
    {
        if (location.getDimensionality() != this.position.getDimensionality())
        {
            throw new RuntimeException("The location provided to WeightedMaxNormHardLocation.similarityWith must have the same dimensionality as the current Hard location.");
        }
        double currentMinimum = Double.MAX_VALUE;
        for (int i = 0; i < location.getDimensionality(); i++)
        {
            double currentValue;
            double currentRadius = ((Double) radius.get(i)).doubleValue();
            if (Math.abs(position.getCoordinate(i) - location.getCoordinate(i)) <= currentRadius)
            {
                currentValue = 1 - Math.abs(position.getCoordinate(i) - location.getCoordinate(i)) 
                						/ currentRadius;
            }
            else
            {
                currentValue = 0;
            }
            if (currentValue < currentMinimum)
            {
                currentMinimum = currentValue;
            }
        }
        return currentMinimum;
    }

    /**
     * Returns the identifier for this hard location.
     * 
     * @return
     */
    public int getId()
    {
        // TODO Auto-generated method stub
        return id;
    } 
}
