/**
 * @file AffineVector.java
 * 
 * @date Aug 7, 2004
 * @author Alexandre Bouchard
 */

package ca.mcgill.rl.util;

/**
 * An interface for elements of an affine finite dimensional real vector space 
 * (approximated by double's). The enclosing vector space can be resized 
 * dynamically.
 */
public interface AffineVector
{
	/**
	 * Performs an addition operation in the standard way.
	 * 
	 * Formally, it should perform v1 := v2 + v1, where v1 is the vector represented
	 * by this. 
	 * 
	 * @param	v2 The vector added to the current one.
	 */
	public void add(AffineVector v2);
	
	/**
	 * Computes the additive inverse.
	 * 
	 * Formally, it should perform v := -v, where v is the vector represented by
	 * this.
	 */
	public void opposite();
	
	/**
	 * Computes a scalar multiplication in the standard way.
	 * 
	 * Formally, it should perform v := scalar * v, where v is the vector
	 * represented by this.
	 * 
	 * @param	scalar The scalar the current vector should be multiplied with.
	 */
	public void times(double scalar);
	
	/**
	 * Accesses one of the components of the vector.
	 * 
	 * @param	index The index of the requested component.
	 */
	public double getComponent(int index);
	
	/**
	 * Sets one of the components of the vector.
	 * 
	 * @param	index The index of the component to set.
	 * @param	value The new value for this component.
	 */
	public void setComponent(int index, double value);
	
	/**
	 * Gets the dimensionality of the enclosing vector space.
	 * 
	 * @return The dimensionality.
	 */
    public int getDimensionality();
    
    /**
     * Resize the enclosing vector space.
     * 
     * If the new dimensionality is higher than the old one, the existing coordinate
     * values will be kept and (new dimensionality - old dimensionality) new
     * coordinates initialized with 0s will be appended. If the new dimensionality is
     * lower, the first newDimensionality coordinates will be kept and the rest will
     * be deleted. Subsequent expansion of the vector space will not retreive the
     * value of these components.
     * 
     * @param 	dimensionality The dimensionality to set.
     */
    public void setDimensionality(int newDimensionality);
}
