/**
 * @file EuclideanLocation.java
 * 
 * @date Aug 7, 2004
 * @author Alexandre Bouchard
 */

package ca.mcgill.rl.util;

import java.util.ArrayList;
import java.util.List;



/**
 * An EuclideanLocation models states of NDP problems with bounded affine, real 
 * valued state spaces.
 * 
 * It uses an ArrayList of double's for the internal representation of the state.
 */
public class EuclideanLocation implements State
{
    /**
     * The representation is an ArrayList.
     */
    private List coordinates;
    
    /**
     * The dimensionality of this location.
     */
    private int dimensionality;
    
    /**
     * Initializes this euclidean location.
     * 
     * Checks that the dimensionality is positive, else throw a RuntimeException.
     * 
     * @param dimensionality The dimensionality of the enclosing vector space.
     */
    public EuclideanLocation(int dimensionality)
    {
        if (dimensionality < 1)
        {
            throw new RuntimeException("The dimensionality of an EuclideanLocation should be positive.");
        }
        this.dimensionality = dimensionality;
        coordinates = new ArrayList(dimensionality);
        for (int i = 0; i < dimensionality; i++)
        {
            coordinates.add(i, null);
        }
    }
   
    /**
     * A copy constructor.
     * 
     * Throw a RuntimeException if the state is null.
     * 
     * @param state
     */
    public EuclideanLocation(EuclideanLocation state)
    {
        if (state == null)
        {
            throw new RuntimeException("Illegal attempt to create an EuclideanLocation from a null state.");
        }
        dimensionality = state.getDimensionality();
        coordinates = new ArrayList(dimensionality);
        
        for (int i = 0; i < state.getDimensionality(); i++)
        {
            coordinates.add(i, new Double(state.getCoordinate(i)));
        }
    }

    /**
     * Gets the value of the i-th coordinate.
     * 
     * Throws a RuntimeException if the coordinate is out of bounds.
     * 
     * @param i The coordinate.
     * @return The value corresponding to this coordinate.
     */    
    public double getCoordinate(int i)
    {
        if (i >= dimensionality || i < 0)
        {
            throw new RuntimeException("Invalid coordinate supplied to EuclideanLocation.getCoordinate().");
        }
        Double coordinate = (Double) coordinates.get(i);
        if (coordinate == null)
        {
            return 0;
        }
        else
        {
            return coordinate.doubleValue();
        }
    }
    
    /**
     * Sets the i-th coordinate of the state to a given value.
     * 
     * The parameter i is assumed to be in the interval [0, dimensionality). Throws
     * a RuntimeException if the coordinate is out of bounds.
     * 
     * @param i The coordinate to set.
     * @param value The new value for this coordinate.
     */
    public void setCoordinate(int i, double value)
    {
        if (i >= dimensionality || i < 0)
        {
            throw new RuntimeException("Invalid coordinate supplied to EuclideanLocation.setCoordinate().");
        }        
        coordinates.set(i, new Double(value));
    }
    
    /**
     * The dimensionality of the enclosing vector space.
     * 
     * It is equal to coordinates.size().
     * 
     * @return
     */
    public int getDimensionality()
    {
        return dimensionality;
    }
}
