/**
 * @file StandardVector.java
 * 
 * @date Aug 8, 2004
 * @author Alexandre Bouchard
 */

package ca.mcgill.rl.util;

import java.util.ArrayList;
import java.util.List;


/**
 * Implements AffineVector in the standard way: an ArrayList from java API. The
 * vector space modelled is finite dimensional, but dynamic resizing of the
 * enclosing vector space is supported.
 */
public class StandardVector implements AffineVector
{
    /**
     * The underlying datastructure to store the vectors is ArrayList.
     */
    private List vectorRepresentation = new ArrayList();
    
    /**
     * The dimensionality of the enclosing vector space.
     */
    private int dimensionality;

    /**
     * Initializes a StandardVector with the given dimensionality for the enclosing
     * vector space.
     */
    public StandardVector(int dimensionality)
    {
        if (dimensionality < 1)
        {
            throw new RuntimeException("Dimensionality of StandardVector must be positive.");
        }
        this.dimensionality = dimensionality;
        for (int i = 0; i < dimensionality; i++)
        {
            vectorRepresentation.add(i, null);
        }
    }    
    
    /**
     * Performs an addition operation in the standard way.
     * 
     * Takes time linear with the dimensionality and assumes that the 
     * dimensionalities are equal. Throws a RuntimeException if the 2 vectors do not
     * have the same dimensionality.
     * 
     * @param	v2 The vector added to the current one.
     */
    public void add(AffineVector v2)
    {
        if (dimensionality != v2.getDimensionality())
        {
            throw new RuntimeException("StandardVectors should have the same dimensionality to be added.");
        }
        for (int i = 0; i < dimensionality; i++)
        { 
            setComponent(i, getComponent(i) + v2.getComponent(i));
        }
    }

    /**
     * Computes the additive inverse.
     * 
     * Takes time linear with the dimensionality.
     */
    public void opposite()
    {
        for (int i = 0; i < getDimensionality(); i++)
        {
            setComponent(i, -getComponent(i));
        }
    }

    /**
     * Computes a scalar multiplication in the standard way.
     * 
     * Takes time linear with the dimensionality.
     * 
     * @param	scalar The scalar the current vector should be multiplied with.
     */
    public void times(double scalar)
    {
        for (int i = 0; i < getDimensionality(); i++)
        {
            setComponent(i, scalar * getComponent(i));
        }
    }

    /**
     * Accesses one of the components of the vector.
     * 
     * Checks that the index is in [0, dimensionality), otherwise throws a 
     * RuntimeException.
     * 
     * @param	index The index of the requested component.
     */
    public double getComponent(int index)
    {
        if (index < 0 || index >= dimensionality)
        {
            throw new RuntimeException("Attempt to access a component of StandardVector that is out of range.");
        }
        Double component = (Double) vectorRepresentation.get(index);
        if (component == null)
        {
            return 0.0;
        }
        else
        {
            return component.doubleValue();
        }
    }

    /**
     * Sets one of the components of the vector.
     * 
     * Checks that the index is in [0, dimensionality), otherwise throws a 
     * RuntimeException.
     * 
     * @param	index The index of the component to set.
     * @param	value The new value for this component.
     */
    public void setComponent(int index, double value)
    {
        if (index < 0 || index >= dimensionality)
        {
            throw new RuntimeException("Attempt to set a component of StandardVector that is out of range.");
        }        
        vectorRepresentation.set(index, new Double(value));
    }

	/**
	 * Gets the dimensionality of the enclosing vector space.
	 * 
	 * @return The dimensionality.
	 */
    public int getDimensionality()
    {
        return dimensionality;
    }

    /**
     * Resize the enclosing vector space.
     * 
     * Checks that the dimensionality is positive, else throw a RuntimeException. If
     * the new dimensionality is higher than the old one, the existing coordinate
     * values will be kept and (new dimensionality - old dimensionality) new
     * coordinates initialized with 0s will be appended. If the new dimensionality is
     * lower, the first newDimensionality coordinates will be kept and the rest will
     * be deleted. Subsequent expansion of the vector space will not retreive the
     * value of these components.
     * 
     * @param 	dimensionality The dimensionality to set.
     */
    public void setDimensionality(int newDimensionality)
    {
        if (dimensionality < 1)
        {
            throw new RuntimeException("Dimensionality of StandardVector must be positive.");
        } 
        if (newDimensionality >= dimensionality)
        {
            for (int i = dimensionality; i < newDimensionality; i++)
            {
                vectorRepresentation.add(i, null);
            }
            this.dimensionality = newDimensionality;
        }
        else
        {
            // Cut the ArrayList to keep only the range of coordinates that is still
            // valid
            vectorRepresentation.subList(newDimensionality, dimensionality).clear();
            this.dimensionality = newDimensionality;
        }
    }
    
    
    /**
     * @return
     */
    public String toString()
    {
        StringBuffer buffer = new StringBuffer();
        for (int i = 0; i < dimensionality; i++)
        {
            buffer.append("" + getComponent(i) + " ");
        }
        return buffer.toString();
    }
}
